/*
 * NN_throught_pass_receiver.cpp
 *
 *  Created on: 2017/07/26
 *      Author: mzmt
 */

#include "NN_through_pass_receiver.h"

using namespace rcsc;
using namespace std;

/*-------------------------------------------------------------------*/
/*! TODO README !!!!
 * パスを行う選手が実行する(ここ重要)、NNスルーパス
 * 計算量の問題があるので、
 * 毎サイクルにRione_libraryで1回だけ行う関数 createThroughtPassKick と
 * SampleFieldEvaluator内で、NNで出た値の参照のみ行う getThroughtPassKickPoint に分けた
 */

/*-------------------------------------------------------------------*/
/*!
 * スルーパスすべき座標を取得する
 * 引数にキッカーを渡す。これは次にボール持つ選手を予想して渡す。
 */
Vector2D
NN_recv::createThroughtPassRecv( const WorldModel & wm , const Vector2D kicker )
{
	// ↓ 入力データ ↓

	// 自分の座標番号
	const int spos = convertPosToFieldNumber( wm.self().pos() );

	// 敵考慮した座標番号
	Vector2D o_middle_pos = getOppMiddlePos( wm, wm.self().pos() );
	if ( o_middle_pos.absX() > 52.5
			|| o_middle_pos.absY() > 34.0)
	{
		return wm.self().pos();
	}
	const int opos = convertPosToFieldNumber( o_middle_pos );


	// パス出す選手の座標番号
	const int kicker_pos = convertPosToFieldNumber( kicker );


	// --------------------------------------------------------
	/* TODO:
	 * ここから NN を使う
	 *
	 * Input1 : kicker_pos ( int型、蹴る選手の座標番号 )
	 * Input2 : spos ( int型、パス受ける自分の座標番号)
	 * Input3 : opos ( int型、sposに近い敵2人の中間地点の座標番号)
	 *
	 * Output : through_pass_pos ( 撃つべきスルーパス座標 )
	 */

	// 出力結果(座標番号)
	const int through_pass_pos = 0; // = NN(Kicker_pos,spos,opos); // XXX ここで計算 XXX

	if ( through_pass_pos <= 0
			|| through_pass_pos > 200 )
	{
		return wm.self().pos();
	}

	// その座標番号のマスの座標の中心あたりを返す
	Vector2D through_pass_point
		= Vector2D( (through_pass_pos / 10) * 5.25 + 2.6 ,	// x座標
					(through_pass_pos % 20) * 3.4 - 1.7 );	// y座標

	return through_pass_point;

}


/*-------------------------------------------------------------------*/
/*!
 * mate_posから近い敵2人 or 敵1人とサイドライン上の点 の中間地点を返す
 */
Vector2D
NN_recv::getOppMiddlePos( const WorldModel & wm, const Vector2D mpos )
{
	// サイドライン上のポイントをもとめる
	Vector2D side_point( mpos.x, mpos.absY() );
	if ( mpos.y < 0 )
		side_point.y *= -1.0;

	// その距離
	double side_dist = mpos.dist( side_point );

	// 欲しい2点の座標
	Vector2D fast = side_point;
	Vector2D second = side_point;

	// 敵ループ
	const PlayerPtrCont::const_iterator opps_end = wm.opponentsFromSelf().end();
	for (PlayerPtrCont::const_iterator o = wm.opponentsFromSelf().begin(); o != opps_end; ++o)
	{
		if( !(*o) ) continue;
		if( (*o)->isGhost() ) continue;

		// 一回目
		if ( fast == second ){
			if ( (*o)->pos().dist( fast ) < (*o)->pos().dist( second ) )
			{
				fast = (*o)->pos();
			}else{
				second = (*o)->pos();
			}
		}

		// それ以降
		if ( fast.dist((*o)->pos()) > mpos.dist((*o)->pos()) )
		{
			second = fast;
			fast = (*o)->pos();
		}
	}

	// 中間地点を返す
	return ( fast + second ) / 2;
}

/*-------------------------------------------------------------------*/
/*!
 * 引数の座標を、フィールド分割した番号に変換する
 * 例外は基本的に0
 */
int
NN_recv::convertPosToFieldNumber( const Vector2D point )
{
	// フィールドの外に出たら
	if ( point.absY() > 34.0
			|| point.x < 0
			|| point.x > 52.5 )
	{
//		printf("ERROR:: NN %d\n",__LINE__);
		return 0;
	}

	// 座標
	double posx = point.x;
	double posy = point.y + 34.0;

	// 縦横で何個目の整数か
	posx = int( posx / 5.25 );
	posy = int( posy / 3.40 ) + 1;

	// 20*10マスの何番にあるか
	int fieldNum = int(posx)*20 + int(posy);

	// 何か事故ったときの例外
	if( fieldNum < 0
			|| fieldNum > 200 )
	{
//		printf("ERROR:: NN %d\n",__LINE__);
		return 0;
	}

	return fieldNum;
}





