/*
 * bhv_wander_move.cpp
 *
 *  Created on: 2017/07/16
 *      Author: mzmt
 */


#include "bhv_wander_move.h"

#include "strategy.h"

#include "bhv_basic_tackle.h"
#include "Rione_library.h"
#include "Rione_opponents_analyzer.h"

#include <rcsc/action/basic_actions.h>
#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/body_intercept.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>

#include <rcsc/player/player_agent.h>
#include <rcsc/player/debug_client.h>
#include <rcsc/player/intercept_table.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>

#include "neck_offensive_intercept_neck.h"
#include "chain_action/field_analyzer.h"

using namespace rcsc;
using namespace std;


/*-------------------------------------------------------------------*/
/*!
 *
 */
bool
Bhv_WanderMove::execute( PlayerAgent * agent )
{
	const ServerParam & SP = ServerParam::i();
	const WorldModel & wm = agent->world();

	// 発動条件のチェック
	if ( !check( agent ) ){
		agent->debugClient().addMessage( "wander check false" );
		return false;
	}

	if ( ! doMove( agent ) )
	{
		agent->debugClient().addMessage( "wander do false" );
		return false;
	}

	return true;
}


/*-------------------------------------------------------------------*/
/*!
	 // #15 さまよいの果て 波は寄せる
 */
bool
Bhv_WanderMove::doMove( PlayerAgent * agent )
{
	const ServerParam & SP = ServerParam::i();
	const WorldModel & wm = agent->world();

	int self_min = wm.interceptTable()->selfReachCycle();
	int mate_min = wm.interceptTable()->teammateReachCycle();
	int opp_min = wm.interceptTable()->opponentReachCycle();
	int our_min = min( self_min, mate_min );
	int reach_min = min( our_min, opp_min );

	Vector2D Bpos = wm.ball().inertiaPoint( our_min );

	Vector2D Spos = Strategy::i().getPosition( wm.self().unum() );

	if ( Spos.dist( wm.self().pos() ) >= 10 )
	{
		return false;
	}

    agent->debugClient().setTarget( Spos );
    agent->debugClient().addCircle( Spos, 0.1 );
    agent->debugClient().addCircle( Spos, 0.5 );
    agent->debugClient().addCircle( Spos, 1.0 );

	int move_time = our_min + int( wm.ball().inertiaPoint( our_min ).dist( Spos )/5 );

	//
	// いい感じにゴールに繋がるパスコースを探したい で、移動したい
	//

	//
	// BposとSposの線分から、受け取り可能なパスコースの三角形を作る
	//
	Segment2D BtoS( Bpos, Spos );
	Line2D BtoSpre = BtoS.line().perpendicular( Spos ); // 自分からの垂線

	Vector2D pos1, pos2;
	if ( Circle2D( Spos + wm.self().vel() , move_time * 0.7 )
			.intersection( BtoS.line().perpendicular( Spos ), &pos1, &pos2 ) != 2 )
	{
		agent->debugClient().addMessage( "circle false" );
		return false;
	}

	agent->debugClient().addTriangle( Bpos, pos1, pos2 );

	//
	// 三角形を左右に分け、敵の少ない方を選ぶ. 同じならゴールに近い方
	//
	Triangle2D tri1( Bpos, Spos, pos1 );
	Triangle2D tri2( Bpos, Spos, pos2 );
	int count1 = 0;
	int count2 = 0;
	const PlayerPtrCont::const_iterator opps_end = wm.opponentsFromSelf().end();
	for (PlayerPtrCont::const_iterator o = wm.opponentsFromSelf().begin(); o != opps_end; ++o)
	{
		if( !(*o) ) continue;
		if( (*o)->isGhost() ) continue;
		if( (*o)->goalie() ) continue;

		if ( tri1.contains( (*o)->pos() ) ) count1++;
		if ( tri2.contains( (*o)->pos() ) ) count2++;
	}

	Triangle2D tri = tri1;
	Vector2D pos;
	if ( count1 > count2 ){
		tri = tri2;
		pos = pos2;
	} else if ( count1 < count2 ){
		tri = tri1;
		pos = pos1;
	} else {
		tri = ( pos1.x > pos2.x ) ? tri1 : tri2;
		pos = ( pos1.x > pos2.x ) ? pos1 : pos2;
	}

	if ( pos.x > wm.ourOffenseLineX() )
		pos.x = wm.ourOffenseLineX();
	//
	// 三角形の頂点に移動
	//
	agent->debugClient().setTarget( pos );
    agent->debugClient().addCircle( pos, 0.5 );

    if ( ! Body_GoToPoint( pos, 0.5, 100
                           ).execute( agent ) )
    {
        Body_TurnToPoint( pos ).execute( agent );
    }
    agent->setNeckAction( new Neck_TurnToBall() );

	agent->debugClient().addMessage( "wander move" );
    return true;
}

/*-------------------------------------------------------------------*/
/*!
 */
bool
Bhv_WanderMove::check( PlayerAgent * agent )
{
	const ServerParam & SP = ServerParam::i();
	const WorldModel & wm = agent->world();

	int self_min = wm.interceptTable()->selfReachCycle();
	int mate_min = wm.interceptTable()->teammateReachCycle();
	int opp_min = wm.interceptTable()->opponentReachCycle();

	int our_min = min( self_min, mate_min );
	int reach_min = min( our_min, opp_min );

	int self_time = opp_min - self_min;		// 自分の余裕サイクル
	int opp_time = our_min - opp_min;		// 相手の余裕サイクル

	const PlayerObject * fast_opp = wm.interceptTable()->fastestOpponent();
	const int fast_opp_unum = ( fast_opp ) ? fast_opp->unum() : -1;

	const PlayerObject * fast_mate = wm.interceptTable()->fastestTeammate();
	const int fast_mate_unum = ( fast_mate ) ? fast_mate->unum() : -1;


	if ( wm.ball().pos().x >= 30
			&& wm.ball().inertiaPoint( our_min ).x > 32
			&& wm.self().pos().x > 35
			&& our_min < opp_min
			&& self_min > mate_min )
	{
		return true;
	}


//	Segment2D BtoSetf( wm.ball().inertiaPoint( our_min ), wm.self().pos() );
//
//	Vector2D right_opp = wm.self().pos();
//	Vector2D left_opp = wm.self().pos();
//	const PlayerPtrCont::const_iterator opps_end = wm.opponentsFromSelf().end();
//	for (PlayerPtrCont::const_iterator o = wm.opponentsFromSelf().begin(); o != opps_end; ++o)
//	{
//		if( !(*o) ) continue;
//		if( (*o)->isGhost() ) continue;
//		if( (*o)->goalie() ) continue;
//
//		if ( left_opp.y > (*o)->pos().y && (*o)->distFromSelf() <= 7 )
//			left_opp = (*o)->pos();
//	}
//
//
    return false;
}



