// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hiroki SHIMORA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 3, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "field_evaluator_basic.h"

#include "NN_through_pass_kicker.h"


#include "lib_bpnn.h"
#include "strategy.h"

#include "field_analyzer.h"
#include "simple_pass_checker.h"
#include "Rione_library.h"
#include "chain_action/cooperative_action.h"

#include <rcsc/player/player_evaluator.h>
#include <rcsc/common/server_param.h>
#include <rcsc/common/logger.h>
#include <rcsc/math_util.h>
#include <rcsc/geom/segment_2d.h>

#include <iostream>
#include <algorithm>
#include <cmath>
#include <cfloat>

#include <stdlib.h>
#include <rcsc/action/body_dribble.h>

// #define DEBUG_PRINT

using namespace rcsc;

static const int VALID_PLAYER_THRESHOLD = 8;
static const double MAX = 100;
static double ten[100][3];
static Vector2D rand_point;



/*-------------------------------------------------------------------*/
/*!

 */
static double evaluate_state( const rcsc::WorldModel & wm, const PredictState & state,
                              const std::vector< ActionStatePair > & path );
static double Half_QF_evaluator( const PredictState & state,
                                 const Vector2D origin,
                                 const Vector2D l_otherpoint,
                                 const Vector2D r_otherpoint );

static double eleven( const WorldModel & wm,
                      const PredictState & state,
                      const std::vector< ActionStatePair > & path );//定義
void randtree(const WorldModel & wm, const PredictState & state);
void cost();
double rrt(const WorldModel & wm, const PredictState & state);
/*-------------------------------------------------------------------*/
/*!

 */
SampleFieldEvaluator::SampleFieldEvaluator()
{

}

/*-------------------------------------------------------------------*/
/*!

 */
SampleFieldEvaluator::~SampleFieldEvaluator()
{

}

/*-------------------------------------------------------------------*/
/*!

 */
double
SampleFieldEvaluator::operator()( const WorldModel & wm, const PredictState & state,
                                  const std::vector< ActionStatePair > & path ) const
{
    const double final_state_evaluation = evaluate_state( wm, state, path );

    double result = final_state_evaluation;

    //
    // ボール保持者が11番だったら
    //
    if( wm.getTeammateNearestToBall( 5, false )
    		&& wm.getTeammateNearestToBall( 5 )->unum() == 11 )
    {
    	result += eleven( wm, state, path );
    }

    return result;
}


/*-------------------------------------------------------------------*/
/*!

 */
static
double
evaluate_state( const WorldModel & wm, const PredictState & state,
		const std::vector< ActionStatePair > & path )
{
	const ServerParam & SP = ServerParam::i();
	const AbstractPlayerObject * holder = state.ballHolder();

	/*
	 * 絶対蹴る or 絶対蹴らないの処理
	 */
	// 捕れる選手がいないとこに蹴らない
	if ( ! holder )
		return - DBL_MAX / 2.0;

	const int holder_unum = holder->unum();

	// 相手ゴールに入るなら蹴る
	if ( state.ball().pos().x > + ( SP.pitchHalfLength() - 0.1 )
			&& state.ball().pos().absY() < SP.goalHalfWidth() + 2.0 )
		return +1.0e+7;

	// 味方ゴールに入るなら蹴らない
	if ( state.ball().pos().x < - ( SP.pitchHalfLength() - 0.1 )
			&& state.ball().pos().absY() < SP.goalHalfWidth() )
		return -1.0e+7;

	// 場外に出るなら蹴らない
	if ( state.ball().pos().absX() > SP.pitchHalfLength()
			|| state.ball().pos().absY() > SP.pitchHalfWidth() )
		return - DBL_MAX / 2.0;

	// キーパーはドリブルしない
	if( wm.self().goalie()
			&& !path.empty()
			&& path[0].M_action->category() == CooperativeAction::Dribble )
	{
		return - DBL_MAX / 2.0;
	}


	//
	// set basic evaluation
	//
	double point = state.ball().pos().x;

	point += std::max( 0.0, (40.0 - ServerParam::i().theirTeamGoalPos().dist( state.ball().pos() ))/2 );

	//
	// 相手が近いならすぐにパスする
	//
	if ( !path.empty()
			&& path[0].M_action->category() == CooperativeAction::Pass
			&& wm.self().unum() <= 9 )
	{
		if ( wm.getDistOpponentNearestToSelf(5,0) < 5 )
			point += 50.0;
	}

	//
	// 相手が近くにいないならホールド
	//
	if ( !path.empty()
			&& path[0].M_action->category() == CooperativeAction::Hold
			&& ( wm.self().unum() >= 7 && wm.self().unum() <= 9 ) )
	{
		if ( wm.getDistOpponentNearestToSelf(5,0) > 5 )
			point += 50.0;
	}

	//
	// 10 11 にパスできても近すぎなら低め
	//
	if ( !path.empty()
			&& path[0].M_action->category() == CooperativeAction::Pass
			&& ( wm.self().unum() >= 7 && wm.self().unum() <= 9 )
			&& holder_unum >= 10
			&& state.ball().pos().x - wm.ball().pos().x < 7 )
	{
		point -= 50.0;
	}

	// XXX
	// 試験導入 NNを用いたスルーパス
	//
	if ( !path.empty()
			&& path[0].M_action->category() == CooperativeAction::Pass )
	{
		Vector2D diff = state.ball().pos() - NN_kick().getNearestNNposFromPoint( wm, state.ball().pos());
		if ( diff != state.ball().pos() )
			if ( diff.absX() <= 2.6
					&& diff.absY() <= 1.7 )
			{
				point += 100.0;
			}
	}

	//
	// 5番 6番 はスルーパスがオフサイドにならないように
	//
	if ( ( wm.self().unum() == 5 || wm.self().unum() == 6 )
			&& holder_unum >= 10
			&& !path.empty()
			&& path[0].M_action->category() == CooperativeAction::Pass )
	{
		if ( holder
				&& holder->isGhost()
				&& holder->posCount() < 2
				&& holder->pos().x > wm.theirDefenseLineX() + holder->pos().dist( wm.self().pos() ) * 0.05 )
		{
			point -= 100;
		}

		if ( wm.ball().pos().x >= 18.0 ){
			point += state.ball().pos().x;
		}

	}


	//
	// クロス時、敵の間を抜けるほどよい
	//
	if ( !path.empty()
			&& path[0].M_action->category() == CooperativeAction::Pass
			&& state.ball().pos().x > 25 )
	{
		double min1 = 1000;
		double min2 = 1000;
		Segment2D pass_seg( wm.ball().pos(), state.ball().pos());
    	const PlayerPtrCont::const_iterator opps_end = state.opponentsFromSelf().end();
    	for (PlayerPtrCont::const_iterator o = state.opponentsFromSelf().begin(); o != opps_end; ++o)
    	{
    		if( !(*o) ) continue;
    		if( pass_seg.dist( (*o)->pos() ) < min1 )
    		{
    			min2 = min1;
    			min1 = pass_seg.dist( (*o)->pos() );
    		}
    	}
    	point *= min1/min2;
	}


	//
	// 味方のフォーメーションの位置に蹴る(下がっている最中にパスミスが起きないように)
	//
	if ( wm.self().unum() <= 6
			&& wm.self().pos().x <= 0 )
	{
		if ( !path.empty()
				&& path[0].M_action->category() == CooperativeAction::Pass )
		{
			point += GravityField_formationPos( wm, state, path , 10.0, -10.0, 0.25, -0.80, 0.0 );
		}
	}

	// ゴール前クリアを安全に
	if ( wm.self().unum() <= 9
			&& wm.self().pos().x < -36)
	{
		if ( !path.empty()
				&& path[0].M_action->category() == CooperativeAction::Pass )
		{
			point += GravityField( wm, state, path , 10.0, -10.0, 0.25, -0.80, 0.0 );
			//        	if ( !wm.ourPlayer(10) && !wm.ourPlayer(11) )
			//        		if ( !wm.ourPlayer(10)->isGhost() && !wm.ourPlayer(11)->isGhost() )
			//        		{
			//        			point += 30 - state.ball().pos().dist( wm.ourPlayer(10)->pos() );
			//        			point += 30 - state.ball().pos().dist( wm.ourPlayer(11)->pos() );
			//        		}
		}
	}

	// TODO
	//	相手の方が向き速度を考えても近い位置には蹴らない
	//

	//
	// PassにのみGravityField適用
	//
	if ( !path.empty()
			&& path[0].M_action->category() == CooperativeAction::Pass
			&& state.ball().pos().x > -10.0
			&& wm.ball().pos().dist( state.ball().pos() ) > 7.0 )
	{
		point += GravityField( wm, state, path , 50.0, -10.0, 0.35, -0.15, 0.0 );
	}
	//
	//    if ( !path.empty()
	//         && path[0].M_action->category() == CooperativeAction::Pass
	//         && wm.self().unum() <= 4
	//		 && wm.ball().pos().x > -30
	//		 )
	//    {
	//    	point += 1000;
	//    }
	//
	//    if ( !path.empty()
	//         && path[0].M_action->category() == CooperativeAction::Dribble
	//		 )
	//    {
	//    	point += 34 - state.ball().pos().absY();
	//    }
	//

	//
	// 過度な蹴鞠パスを防止する
	//
	point *= ReturnPass( wm, state, path , 0.2 );


	//
	// 味方近距離へのドリブルを防止する
	//
	point += DribbleToTeammate( wm, state, path , -30.0 );


	//
	// 六番を原点とした二変数関数評価値
	//
	if ( !path.empty()
			&& path[0].M_action->category() == CooperativeAction::Pass
			&& wm.ball().pos().x > 28.0
			&& wm.ourPlayer( 9 )
			&& wm.ourPlayer( 7 )
			&& wm.ourPlayer( 8 ) )
	{
		point += Half_QF_evaluator( state,
				wm.ourPlayer( 9 )->pos(),
				wm.ourPlayer( 7 )->pos(),
				wm.ourPlayer( 8 )->pos() ) * 0.95;
	}

	if ( !path.empty()
			&& path[0].M_action->category() == CooperativeAction::Pass
			&& wm.ball().pos().x > 28.0
			&& wm.ourPlayer( int((wm.ball().pos().y<0) ? 5 : 6) )
			&& wm.ourPlayer( int((wm.ball().pos().y<0) ? 7 : 8) )
			&& wm.ourPlayer( 9 ) )
	{
		point += Half_QF_evaluator( state,
				wm.ourPlayer( int((wm.ball().pos().y<0) ? 5 : 6) )->pos(),
				wm.ourPlayer( int((wm.ball().pos().y<0) ? 7 : 8) )->pos(),
				wm.ourPlayer( 9 )->pos() ) * 0.70;
	}

	//
	// ドリブルに対する評価値
	//
	point += SafeDribble( wm, state, path );


//	// TODO
//	// 前方に余裕があるときのドリブル
//	//
//    const Vector2D arrow = ( state.ball().pos() - wm.self().pos() ) * 3.0;
//    const Vector2D goal_pos(52.5, 0); // ゴール
//
//    if ( wm.self().unum() >= 7
//    		&& wm.self().unum() <= 9
//    		&& -25.0 < state.ball().pos().x
//    		&& state.ball().pos().x < 40.0 && arrow.x > 0 )
//    {
//    	const AngleDeg myangle = (wm.self().pos() - state.ball().pos()).th();
//    	const Sector2D sec(wm.self().pos(), 10.0, 14.5, myangle - 30.0, myangle + 30.0); // おうぎ型
//    	const PlayerPtrCont::const_iterator opps_end = state.opponentsFromSelf().end();
//
//    	for (PlayerPtrCont::const_iterator o = state.opponentsFromSelf().begin(); o != opps_end; ++o)
//    	{
//    		if( !(*o) ) continue;
//    		if( !sec.contains( (*o)->pos() )) // ドリブル方向に相手がいなかったら
//    		{
//    			if ( !path.empty()
//    					&& path[0].M_action->category() == CooperativeAction::Dribble ){
//    				point += std::max( 0.0, arrow.x * 10.0) - goal_pos.dist(state.ball().pos());
//    			}
//    		} else {
//    			break;
//    		}
//    	}
//    }


    // TODO
    // フォワードは無理やりでもドリブルして前にもっていく
    //
	if ( wm.self().unum() >= 10
			&& wm.self().pos().x >= 10.0
			&& wm.self().pos().absY() >= 25
			&& wm.ball().pos().x < state.ball().pos().x
			&& !path.empty()
			&& path[0].M_action->category() == CooperativeAction::Dribble )
	{
		point *= 1.5;
	}

	//
	// コーナーから少し戻したあたりの高さを想定
	//
	if ( ( wm.self().unum() >= 5 &&  wm.self().unum() <= 9 )
			&& ( wm.self().pos().x >= 25.0 && wm.self().pos().x <= 42.0 )
			&& !path.empty()
			&& path[0].M_action->category() == CooperativeAction::Pass )
	{
		// サイドからは内側に通す
		if ( wm.self().pos().absY() >= 15
				&& wm.ball().pos().absY() - 4 > state.ball().pos().absY()
				&& ( state.ball().pos().x >= 25.0 && state.ball().pos().x <= 42.0 ) )
		{
			point += 30;
			point *= 1.5;
		}

		// 内側にいるときは11番などに通す
		if ( wm.self().pos().absY() <= 11
				&& state.ball().pos().x - wm.ball().pos().x > 5
				&& holder_unum >= 9 )
		{
			point += 50;
			point *= 1.8;
		}

	}


	//
	// スルーパスに強制的に加点
	//
	if( !path.empty()
			&& ( path[0].M_action->description() == "strictThrough"
					|| path[0].M_action->description() == "rioneThrough_N"
					|| path[0].M_action->description() == "rioneThrough" ))
	{
		point += std::max( 0.0, 100.0 - ServerParam::i().theirTeamGoalPos().dist( state.ball().pos() )) * 0.2;
	}

	//
	// パス相手の座標の信用度
	//
	point *= std::pow( 0.98, state.ballHolder()->posCount() );


	// シュートチャンスな状況
	if ( FieldAnalyzer::can_shoot_from
			( holder->unum() == state.self().unum(),
					holder->pos(),
					state.getPlayerCont( new OpponentOrUnknownPlayerPredicate( state.ourSide() ) ),
					VALID_PLAYER_THRESHOLD ) )
	{
		point += 1.0e+6;
		if ( holder_unum == state.self().unum() )
			point += 5.0e+5;
	}

	return point;
}


/*-------------------------------------------------------------------*/
/*!
 * @brief originを原点とした、r,l_otherpointを通る二次関数を元にした2変数関数
 * goal前でボールを回すために。
 * @param state
 * @param origin 原点の選手の座標(6番)
 * @param l_otherpoint 左側の交点とする選手の座標（7番）
 * @param r_otherpoint 右側の交点とする選手の座標（8番）
 * @return 付加するpoint
 */
static
double
Half_QF_evaluator( const PredictState & state,
                   const Vector2D origin,
                   const Vector2D l_otherpoint,
                   const Vector2D r_otherpoint )
{

    double point = 0.0;
    double x, y, a;
    Vector2D next_pos = state.ball().pos();
    double dx = next_pos.x - origin.x;
    double dy = next_pos.y - origin.y;

    if( origin.y > l_otherpoint.y && l_otherpoint.x >= origin.x )
    {
        x = origin.y - l_otherpoint.y;
        y = l_otherpoint.x - origin.x;
        a = y / ( x * x );
        if( dx >= a * dy * dy )
        {
        	point += std::max( 100 - std::sqrt( dy * dy + ( a * dy * dy ) * ( a * dy * dy ) ), 0.0 );

            return point;
        }
    }

    if( origin.y < r_otherpoint.y && r_otherpoint.x >= origin.x)
    {
        x = r_otherpoint.y - origin.y;
        y = r_otherpoint.x - origin.x;
        a = y / ( x * x );
        if( dx >= a * dy * dy )
        {
        	point += std::max( 100 - std::sqrt( dy * dy + ( a * dy * dy ) * ( a * dy * dy ) ), 0.0 );

            return point;
        }
    }

    return point;
}

/*-------------------------------------------------------------------*/
//11番の行動
/*-------------------------------------------------------------------*/
double eleven( const WorldModel & wm,
			   const PredictState & state,
			   const std::vector< ActionStatePair > & path)
{
	const AbstractPlayerObject * holder = state.ballHolder();
	double point = 0;
	int i = 0;

	if(!path.empty())
	{
		for(i = 0; i < path.size() - 1; i++)
		{
			if( path[i].state().ballHolderUnum() == 11 )
				if( path[i+1].state().ballHolderUnum() == 9
						|| path[i+1].state().ballHolderUnum() == 10 )
					point += 10.0;
		}
	}

	// パスしないようにドリブル加算
	if ( !path.empty()
			&& path[0].M_action->category() == CooperativeAction::Dribble)
	{
		point += 5.0;	//50以下 maxが元々400まで 5 or 10
	}

	//パスの評価値を前・横寄りに。 9 or 10 番号指定してパス 選手間の距離も考慮
	//パス対象選手とボールの位置の差 state.ballHolderUnum()はパス相手の背番号
	if(state.ballHolderUnum() == 9 || state.ballHolderUnum() == 10)
	{
		if( wm.getTeammateNearestToBall( 5, false ) )
			if( wm.getTeammateNearestToBall( 5 )->unum() == 11 )
				point += 5.0;	// 11 -> 10 or 9 のパスに加算
	}
	return point;
}



/*-------------------------------------------------------------------*/
// GravityFieldの関数
// Max,Min 最大値,最小値
// OurPlayer,OppPlayer,SelfGravity 反映する割合
/*-------------------------------------------------------------------*/
double GravityField( const rcsc::WorldModel & wm,
					 const PredictState & state,
					 const std::vector< ActionStatePair > & path ,
					 double Max, double Min,
					 double OurPlayer, double OppPlayer, double SelfGravity )
{
	const double G = 6.67;			// 万有引力定数
	const double opp_M = 13.423767;		// エージェントの質量を設定
	const double our_M = 15.336398;
	Vector2D target_point = state.ball().pos();
	double point = 0;				//評価値

	const PlayerPtrCont::const_iterator opps_end = state.opponentsFromSelf().end();
	for (PlayerPtrCont::const_iterator o = state.opponentsFromSelf().begin(); o != opps_end; ++o)
	{
		if( !(*o) ) continue;
		double for_divide = 100 - ServerParam::i().ourTeamGoalPos().dist(state.self().pos());
		double point_max = ( G * (for_divide * 0.01 * opp_M) )
                        		/ ( (*o)->pos().dist(target_point) * (*o)->pos().dist(target_point) );

		point += std::max( 0.0, point_max ) * OppPlayer; // 相手の評価値
	}

	const PlayerPtrCont::const_iterator mates_end = state.teammatesFromSelf().end();
	for (PlayerPtrCont::const_iterator o = state.teammatesFromSelf().begin(); o != mates_end; ++o)
	{
		if( !(*o) ) continue;
		double for_divide = 100 - ServerParam::i().theirTeamGoalPos().dist(state.self().pos());
		double point_max = ( G * (for_divide * 0.01 * our_M) )
                        		/ ( (*o)->pos().dist(target_point) * (*o)->pos().dist(target_point) );

		if ( wm.self().unum() == (*o)->unum() )
			point += std::max( 0.0, point_max) * SelfGravity; // 自身の評価値
		else if ( (*o)->pos() != wm.self().pos() )
			point += std::max( 0.0, point_max) * OurPlayer; // 味方の評価値
	}

	return std::min(std::max(point, Min), Max); // 最大最小の範囲内に調整
}
/*-------------------------------------------------------------------*/
// GravityFieldの関数
// Max,Min 最大値,最小値
// OurPlayer,OppPlayer,SelfGravity 反映する割合
/*-------------------------------------------------------------------*/
double GravityField_formationPos( const rcsc::WorldModel & wm,
					 const PredictState & state,
					 const std::vector< ActionStatePair > & path ,
					 double Max, double Min,
					 double OurPlayer, double OppPlayer, double SelfGravity )
{
	const double G = 6.67;			// 万有引力定数
	const double opp_M = 13.423767;		// エージェントの質量を設定
	const double our_M = 15.336398;
	Vector2D target_point = state.ball().pos();
	double point = 0;				//評価値

	const PlayerPtrCont::const_iterator opps_end = state.opponentsFromSelf().end();
	for (PlayerPtrCont::const_iterator o = state.opponentsFromSelf().begin(); o != opps_end; ++o)
	{
		if( !(*o) ) continue;
		double for_divide = 100 - ServerParam::i().ourTeamGoalPos().dist(state.self().pos());
		double point_max = ( G * (for_divide * 0.01 * opp_M) )
                        		/ ( (*o)->pos().dist(target_point) * (*o)->pos().dist(target_point) );

		point += std::max( 0.0, point_max ) * OppPlayer; // 相手の評価値
	}

	for ( int unum = 1; unum <= 11; unum++ ){
		Vector2D opos = Strategy::i().getPosition( unum );
		double for_divide = 100 - ServerParam::i().theirTeamGoalPos().dist(state.self().pos());
		double point_max = ( G * (for_divide * 0.01 * our_M) )
                        		/ ( opos.dist(target_point) * opos.dist(target_point) );

		if ( wm.self().unum() == unum )
			point += std::max( 0.0, point_max) * SelfGravity; // 自身の評価値
		else if ( opos != wm.self().pos() )
			point += std::max( 0.0, point_max) * OurPlayer; // 味方の評価値

	}

	return std::min(std::max(point, Min), Max); // 最大最小の範囲内に調整
}

/*-------------------------------------------------------------------*/
// 過度な蹴鞠パスを防止する関数
// パス元に返すパスの場合は Rate を それ以外は 1 を返す
/*-------------------------------------------------------------------*/
double ReturnPass( const rcsc::WorldModel & wm,
				   const PredictState & state,
				   const std::vector< ActionStatePair > & path ,
			       double Rate )
{
    if ( !path.empty()
         && path[0].M_action->category() == CooperativeAction::Pass )
    {
        const PlayerPtrCont::const_iterator mates_end = state.teammatesFromSelf().end();
        for ( PlayerPtrCont::const_iterator o = state.teammatesFromSelf().begin(); o != mates_end; ++o)
        {
            int last_kick_player = Unum_Unknown;

            if( (*o)->kicked() )
            	last_kick_player = (*o)->unum();

            if( state.ballHolderUnum() == last_kick_player )
                return Rate;
        }
    }
	return 1.0;
}

/*-------------------------------------------------------------------*/
// 味方近距離へのドリブルを防止する関数
// Point の値を人数分だけ返す
/*-------------------------------------------------------------------*/
double DribbleToTeammate( const rcsc::WorldModel & wm,
						  const PredictState & state,
						  const std::vector< ActionStatePair > & path ,
						  double Point )
{
    const ServerParam & SP = ServerParam::i();
	double p = 0.0;
    if ( !path.empty()
         && path[0].M_action->category() == CooperativeAction::Dribble
         && ! SP.theirPenaltyArea().contains( state.ball().pos() ) )
    {
        const PlayerPtrCont::const_iterator mates_end = state.teammatesFromSelf().end();
        for (PlayerPtrCont::const_iterator t = state.teammatesFromSelf().begin();
             t != mates_end; ++t)
        {
            if ( !(*t) ) continue;
            if ( (*t)->distFromSelf() > 10.0 ) continue;

            // これ以上近づいちゃダメな距離
            double dribble_eprohibited_distance =  (*t)->playerTypePtr()->kickableArea()
                                                   + (*t)->playerTypePtr()->playerSpeedMax() * state.spendTime();
            if ( (*t)->pos().dist( state.ball().pos() ) < dribble_eprohibited_distance )
            {
                p += Point;
            }
        }
    }
	return p;
}

/*-------------------------------------------------------------------*/
// 前方へのドリブルを行う関数
// 敵が多いほど加算値は少ない
/*-------------------------------------------------------------------*/
double SafeDribble( const rcsc::WorldModel & wm,
		  	  	  const PredictState & state,
				  const std::vector< ActionStatePair > & path )
{
	double point = 0;
    const Rect2D l_cross( Vector2D( 40.0, -20.0 ), Vector2D( 52.5, -34.0) );
    const Rect2D r_cross( Vector2D( 40.0, 34.0 ), Vector2D( 52.5, 20.0 ) );

    for ( size_t i = 0; i < path.size(); ++i )
    {
        if ( !path.empty()
             && path[i].M_action->category() == CooperativeAction::Dribble
             && state.ball().pos().x > 0.0 && wm.ball().pos().x < 40.0
             && ( ! r_cross.contains( state.ball().pos() ) || ! l_cross.contains( state.ball().pos() ) )
             && state.ballHolderUnum() > 8 )
        {
            // フォワードのエージェントがどれだけ前に居るか
            double f_dist = 0.0;

            const PlayerPtrCont::const_iterator opps_end = state.opponentsFromSelf().end();
            for (PlayerPtrCont::const_iterator o = state.opponentsFromSelf().begin();
                 o != opps_end; ++o)
            {
                if( (*o)->goalie() ) continue;

                // 前にいる敵の距離*人数分だけ加算
                f_dist += std::max( 0.0, (*o)->pos().x - wm.self().pos().x );
            }

            //ドリブルに付加する値
            double dribble_point = std::max( 0.0,
            								120.0 - ServerParam::i().theirTeamGoalPos().dist( state.ball().pos() ) - f_dist );

            //オフサイドライン近く && 前方向のドリブルのみ
            if( wm.self().pos().x > wm.offsideLineX() - 3.0
                && state.ball().pos().x > wm.self().pos().x )
            {
                point +=  dribble_point + std::max( 0.0, wm.self().pos().x - 36.0 ) * 0.5;
            }
        }
    }

	return point;
}

/*-------------------------------------------------------------------*/
// TDP2016 RRT
/*-------------------------------------------------------------------*/
void randtree(const WorldModel & wm, const PredictState & state){
    int a = 0;


    while(a < MAX){
        bool flag = false;
        rand_point.x = 0.0+(double)rand()*(52.5-0+1.0)/(1.0+RAND_MAX);
        ten[a][0] = rand_point.x;
        rand_point.y = -34.0+(double)rand()*(34.0-(-34.0)+1.0)/(1.0+RAND_MAX);
        ten[a][1] = rand_point.y;
//	  dlog.addText( Logger::ACTION_CHAIN,
//			"rand_point %f, %f", rand_point.x, rand_point.y );

        int count=0;
        const PlayerPtrCont::const_iterator opps_end = state.opponentsFromSelf().end();
        for (PlayerPtrCont::const_iterator o = state.opponentsFromSelf().begin();
             o != opps_end; ++o){
            if((*o)->pos().dist(rand_point) <= 3){
                flag = true;
            }

            Line2D line( rand_point, wm.self().pos() );
            Circle2D opp_circle( (*o)->pos(), 3 );
            Vector2D pos1, pos2;
            count = opp_circle.intersection(line, &pos1, &pos2);
        }

        if(flag || count >= 1){
            continue;
        }
        a++;
    }
}

void cost(){
    Vector2D goal_point;
    goal_point.x = 36.0; //目標地点のx座標
    double x_num = 100.0; //適当な実数値
    int i = 0;

    while(i < MAX){
        ten[i][2] = x_num - (goal_point.x - rand_point.x);
        i++;
    }
}

double rrt(const WorldModel & wm, const PredictState & state){
    Vector2D target_point;
    int i = 0;
    int j = 0;
    double *d;
    int num = 0;
    double angle_c;
    double result = 0;


    while(1){
        randtree(wm, state);
        if(rand_point.x > 36.0 || ( -20 < rand_point.y && rand_point.y < 20 )){
            break;
        }
        cost();

        for(i = 0; i < MAX-1; i++){
            for(j = MAX-1; j > i; j--){
                if(ten[j-1][2] < ten[j][2]){
                    d = ten[j-1];
                    *ten[j-1] = *ten[j];
                    *ten[j] = *d;
                }
            }
        }

        target_point.x = ten[0][0];
        target_point.y = ten[0][1];

        dlog.addText( Logger::ACTION_CHAIN,
                      "target_point %f, %f", target_point.x, target_point.y );


        AngleDeg target_angle = (target_point - wm.self().pos()).th();
        AngleDeg ball_angle = (state.ball().pos() - wm.self().pos()).th();

        angle_c = target_angle.degree() - ball_angle.degree();

        if(angle_c <= 5){           //target_angleとball_angleの差が一番小さい時
            result += 100;
        }
    }

    return result;
}
